if bit == nil then bit = {} end

bit.MEMORY_HEIGHT = 64
bit.MEMORY_WIDTH = 64
bit.PC_ADDR = 4088 -- program counter address (2 bytes)
bit.CARRY_ADDR = 4090 -- carry bit address (smallest bit)
bit.INTERRUPT_ADDR = 4091 -- interrupt address (manual input)
bit.CLOCK_ADDR = 4092 -- system clock address (4 bytes)

bit.CLEAR = 0
bit.CLEARREG = 0
bit.SAVE = 0
bit.LOAD = 0

function bit.cycle()

	local pc = bit.readByte(bit.PC_ADDR) * 256 + bit.readByte(bit.PC_ADDR + 1)
	local cloc = bit.readByte(bit.CLOCK_ADDR + 3) + 1
	
	-- loops are for wussies
	if(cloc == 256) then
		bit.writeByte(bit.CLOCK_ADDR + 3, 0)
		local cloc2 = bit.readByte(bit.CLOCK_ADDR + 2) + 1
		if(cloc2 == 256) then
			bit.writeByte(bit.CLOCK_ADDR + 2, cloc2)
			cloc2 = bit.readByte(bit.CLOCK_ADDR + 1) + 1
			if(cloc2 == 256) then
				bit.writeByte(bit.CLOCK_ADDR + 1, cloc2)
				cloc2 = bit.readByte(bit.CLOCK_ADDR) + 1
				if(cloc2 == 256) then
					bit.writeByte(bit.CLOCK_ADDR, cloc2)
				else
					bit.writeByte(bit.CLOCK_ADDR, cloc2)
				end
			else
				bit.writeByte(bit.CLOCK_ADDR + 1, cloc2)
			end
		else
			bit.writeByte(bit.CLOCK_ADDR + 2, cloc2)
		end
	else
		bit.writeByte(bit.CLOCK_ADDR + 3, cloc)
	end
	
	local op = bit.readByte(pc)
	
	--msg("op-code: "..op)
	
	-- 0 is NOP
	if(op == 1) then -- write a value to a register
		bit.writeRegister(bit.readByte(pc + 1), bit.readByte(pc + 2))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 3) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 3) % 256)
	elseif(op == 2) then -- copy a value from a register to another (registers are stored as half-bytes)
		bit.copyRegister(bit.readByte(pc + 1))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 2) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 2) % 256)
	elseif(op == 3) then -- copy from a register to memory
		bit.writeRegMem(bit.readByte(pc + 1), (bit.readByte(pc + 2) * 16 + bit.readByte(pc + 3)))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 4) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 4) % 256)
	elseif(op == 4) then -- copy a value from memory to a register
		bit.writeMemReg((bit.readByte(pc + 1) * 16 + bit.readByte(pc + 2)), bit.readByte(pc + 3))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 4) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 4) % 256)
	elseif(op == 5) then -- jump pc to addr
		bit.writeByte(bit.PC_ADDR, bit.readByte(pc + 1))
		bit.writeByte(bit.PC_ADDR + 1, bit.readByte(pc + 2))
	elseif(op == 6) then -- jump pc to addr if reg1 = reg2
		if(bit.equalRegister(bit.readByte(pc + 1))) then
			bit.writeByte(bit.PC_ADDR, bit.readByte(pc + 2))
			bit.writeByte(bit.PC_ADDR + 1, bit.readByte(pc + 3))
		else
			bit.writeByte(bit.PC_ADDR, math.floor((pc + 4) / 256))
			bit.writeByte(bit.PC_ADDR + 1, (pc + 4) % 256)
		end
	elseif(op == 7) then -- jump pc to addr if reg1 < reg2
		if(bit.lessThanRegister(bit.readByte(pc + 1))) then
			bit.writeByte(bit.PC_ADDR, bit.readByte(pc + 2))
			bit.writeByte(bit.PC_ADDR + 1, bit.readByte(pc + 3))
		else
			bit.writeByte(bit.PC_ADDR, math.floor((pc + 4) / 256))
			bit.writeByte(bit.PC_ADDR + 1, (pc + 4) % 256)
		end
	elseif(op == 8) then -- add the value of the second register to the first
		bit.addRegisters((bit.readByte(pc + 1)))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 2) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 2) % 256)
	elseif(op == 9) then -- subtract the value of the second register from the first
		bit.subRegisters((bit.readByte(pc + 1)))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 2) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 2) % 256)
	elseif(op == 10) then -- multiply the value of the first register with the second
		bit.multRegisters((bit.readByte(pc + 1)))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 2) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 2) % 256)
	elseif(op == 11) then -- divide the value of the first register by the second, remainder stored in the second
		bit.divRegisters((bit.readByte(pc + 1)))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 2) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 2) % 256)
	elseif(op == 12) then -- bitwise and the second register onto the first
		bit.andRegisters((bit.readByte(pc + 1)))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 2) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 2) % 256)
	elseif(op == 13) then -- bitwise or the second register onto the first
		bit.orRegisters((bit.readByte(pc + 1)))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 2) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 2) % 256)
	elseif(op == 14) then -- bitwise not two registers (one if both registers are equal)
		bit.notRegisters((bit.readByte(pc + 1)))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 2) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 2) % 256)
	elseif(op == 15) then -- bitwise xor the second register onto the first
		bit.notRegisters((bit.readByte(pc + 1)))
		bit.writeByte(bit.PC_ADDR, math.floor((pc + 2) / 256))
		bit.writeByte(bit.PC_ADDR + 1, (pc + 2) % 256)
	end

end

function bit.setInput()

	local val = 0

	if(not entity(60, 64, "state")) then val = val + 128 end
	if(not entity(61, 64, "state")) then val = val + 64 end
	if(not entity(62, 64, "state")) then val = val + 32 end
	if(not entity(63, 64, "state")) then val = val + 16 end
	if(not entity(65, 64, "state")) then val = val + 8 end
	if(not entity(66, 64, "state")) then val = val + 4 end
	if(not entity(67, 64, "state")) then val = val + 2 end
	if(not entity(68, 64, "state")) then val = val + 1 end
	
	bit.writeByte(bit.INTERRUPT_ADDR, val)

end

function bit.handleTrigger(name,src)

	if(name == "exec") then
		addhook("ms100","bit.cycle")
		msg("excection started!")
	elseif(name == "abort") then
		freehook("ms100","bit.cycle")
		msg("excection aborted!")
	elseif(name == "enter") then
		bit.setInput()
		msg("input set!")
	elseif(name == "clear") then
		bit.trigClear()
	elseif(name == "clearReg") then
		bit.trigClearReg()
	elseif(name == "save") then
		bit.trigSave()
	elseif(name == "load") then
		bit.trigLoad()
	end

end

function bit.handleFKeys(id,action)

	if(action == 1) then
		addhook("ms100","bit.cycle")
		msg("excection started!")
	elseif(action == 2) then
		freehook("ms100","bit.cycle")
		msg("excection aborted!")
	elseif(action == 3) then
		bit.setInput()
		msg("input set!")
	end

end

function bit.trigClear()

	if(bit.CLEAR == 0) then
		bit.CLEAR = 1
		msg("clear all buildings?")
		timer(5000,"bit.resetClear")
	elseif(bit.CLEAR > 0) then
		bit.CLEAR = 0
		msg("clearing all buildings...")
		freetimer("bit.resetClear")
		
		local objectlist=object(0,"table")
		for _,id in pairs(objectlist) do
		   parse("killobject "..id)
		end
		
		msg("All buildings have been cleared!")
	end

end

function bit.resetClear()

	freetimer("bit.resetClear")
	bit.CLEAR = 0
	msg("building clearing has been aborted.")

end

function bit.trigClearReg()

	if(bit.CLEARREG == 0) then
		bit.CLEARREG = 1
		msg("clear all registers?")
		timer(5000,"bit.resetClearReg")
	elseif(bit.CLEARREG > 0) then
		bit.CLEARREG = 0
		msg("clearing all registers...")
		freetimer("bit.resetClearReg")
		
		local objectlist=closeobjects((70 * 32), (52 * 32), 128)
		for _,id in pairs(objectlist) do
		   parse("killobject "..id)
		end
		
		local objectlist2=closeobjects((70 * 32), (75 * 32), 128)
		for _,id in pairs(objectlist2) do
		   parse("killobject "..id)
		end
		
		msg("All registers have been cleared!")
	end

end

function bit.resetClearReg()

	freetimer("bit.resetClearReg")
	bit.CLEARREG = 0
	msg("registers clearing has been aborted.")

end

function bit.trigSave()

	if(bit.SAVE == 0) then
		bit.SAVE = 1
		msg("Save the program to file?")
		timer(5000,"bit.resetSave")
	elseif(bit.SAVE > 0) then
		bit.SAVE = 0
		msg("Saving the program to PROGRAM.CS2D-BIN...")
		freetimer("bit.resetSave")
		
		io.output("sys/lua/PROGRAM.CS2D-BIN")
		
		local data = ""
		
		for i=0,(bit.MEMORY_WIDTH-1)*2,2 do
		
			for j=0,bit.MEMORY_HEIGHT-1,1 do
			
				local str = string.sub(string.format("%x", bit.readBytexy(i,j)),1,2)
				
				if(string.len(str) < 2) then str = "0"..str end
			
				data = data..str
			
			end
			
			data = data.."\n"
		
		end
		
		io.write(data)
		io.close()
		
		msg("Successfully saved!")
		
	end

end

function bit.resetSave()

	freetimer("bit.resetSave")
	bit.SAVE = 0
	msg("Saving has been aborted.")

end

function bit.trigLoad()

	if(bit.LOAD == 0) then
		bit.LOAD = 1
		msg("Load a program from file?")
		timer(5000,"bit.resetLoad")
	elseif(bit.LOAD > 0) then
		bit.LOAD = 0
		msg("Loading the program from PROGRAM.CS2D-BIN...")
		freetimer("bit.resetLoad")
		
		local i, byt = 0, ""
		
		for line in io.lines("sys/lua/PROGRAM.CS2D-BIN") do
		
			for j=0,(string.len(line)-1)*2,2 do
			
				local byt = string.sub(line, j + 1, j + 2)
			
				if(string.len(byt) == 2) then
					bit.writeBytexy(i*2, j/2, tonumber(byt,16))
				end
				
			
			end
				
			i = i + 1
		
		end
		
		msg("Successfully loaded!")
	end

end

function bit.resetLoad()

	freetimer("bit.resetLoad")
	bit.LOAD = 0
	msg("Loading has been aborted.")

end


-- operations


function bit.writeRegister(reg, val) -- write value to register

	bit.writeBytexy(bit.registerX(reg), bit.registerY(reg), val)

end


function bit.copyRegister(reg) -- copy value of regA to regB

	local regA, regB = math.floor(reg / 16), reg % 16

	bit.writeBytexy(bit.registerX(regB), bit.registerY(regB), bit.readBytexy(bit.registerX(regA), bit.registerY(regA)))

end

function bit.writeRegMem(reg, addr) -- write value from register to memory

	bit.writeByte(bit.readBytexy(bit.registerX(reg), bit.registerY(reg)))

end

function bit.writeRegMem(addr, reg) -- write value from memory to register

	bit.writeBytexy(bit.registerX(reg), bit.registerY(reg), bit.readByte(addr))

end

function bit.equalRegister(reg) -- returns true if regA and regB have equal values 

	local regA, regB = math.floor(reg / 16), reg % 16

	return bit.readBytexy(bit.registerX(regA), bit.registerY(regA)) == bit.readBytexy(bit.registerX(regB), bit.registerY(regB))

end

function bit.lessRegister(reg) -- returns true if the value of regA is smaller than that of regB

	local regA, regB = math.floor(reg / 16), reg % 16

	return bit.readBytexy(bit.registerX(regA), bit.registerY(regA)) < bit.readBytexy(bit.registerX(regB), bit.registerY(regB))

end

function bit.addRegisters(reg) -- adds the value of regB to regA

	local regA, regB = math.floor(reg / 16), reg % 16
	
	local valA, valB = bit.readBytexy(bit.registerX(regA), bit.registerY(regA)), bit.readBytexy(bit.registerX(regB), bit.registerY(regB))
	
	local valC = valA + valB
	
	if(valC > 255) then 
		valC = valC % 256
		bit.writeByte(bit.CARRY_ADDR, 1)
	else
		bit.writeByte(bit.CARRY_ADDR, 0)
	end
	
	bit.writeBytexy(bit.registerX(regA), bit.registerY(regA), valC)

end

function bit.subRegisters(reg) -- subtracts the value of regB from regA

	local regA, regB = math.floor(reg / 16), reg % 16
	
	local valA, valB = bit.readBytexy(bit.registerX(regA), bit.registerY(regA)), bit.readBytexy(bit.registerX(regB), bit.registerY(regB))
	
	local valC = valA - valB
	
	if(valC < 0) then 
		valC = valC % 256
		bit.writeByte(bit.CARRY_ADDR, 1)
	else
		bit.writeByte(bit.CARRY_ADDR, 0)
	end
	
	bit.writeBytexy(bit.registerX(regA), bit.registerY(regA), valC)

end

function bit.multRegisters(reg) -- multiplies the value of regB to regA, using regB for higher bits

	local regA, regB = math.floor(reg / 16), reg % 16
	
	local valA, valB = bit.readBytexy(bit.registerX(regA), bit.registerY(regA)), bit.readBytexy(bit.registerX(regB), bit.registerY(regB))
	
	local valC = valA * valB
	
	bit.writeBytexy(bit.registerX(regB), bit.registerY(regB), math.floor((valA * valB) / 256))
	bit.writeBytexy(bit.registerX(regA), bit.registerY(regA), valC % 256)

end

function bit.divRegisters(reg) -- divides the value of regA by regB, saves the remainder to regB

	local regA, regB = math.floor(reg / 16), reg % 16
	
	local valA, valB = bit.readBytexy(bit.registerX(regA), bit.registerY(regA)), bit.readBytexy(bit.registerX(regB), bit.registerY(regB))
	
	bit.writeBytexy(bit.registerX(regA), bit.registerY(regA), valA / valB)
	bit.writeBytexy(bit.registerX(regB), bit.registerY(regB), valA % valB)

end

function bit.andRegisters(reg) -- bitwise and the value of regB to regA

	local regA, regB = math.floor(reg / 16), reg % 16
	
	local valA, valB = bit.readBytexy(bit.registerX(regA), bit.registerY(regA)), bit.readBytexy(bit.registerX(regB), bit.registerY(regB))
	
	local c = 0
	
	for i=7,0,-1 do
	
		local a0, b0 = valA - 2^i >= 0, valB - 2^i >= 0
		
		if(a0 and b0) then c = c + 2^i end
		if(a0) then valA = valA - 2^i end
		if(b0) then valB = valB - 2^i end
		if(valA == 0) then break end
		if(valB == 0) then break end
	
	end
	
	bit.writeBytexy(bit.registerX(regA), bit.registerY(regA), c)

end

function bit.orRegisters(reg) -- bitwise or the value of regB to regA

	local regA, regB = math.floor(reg / 16), reg % 16
	
	local valA, valB = bit.readBytexy(bit.registerX(regA), bit.registerY(regA)), bit.readBytexy(bit.registerX(regB), bit.registerY(regB))
	
	local c = 0
	
	for i=7,0,-1 do
	
		local a0, b0 = valA - 2^i >= 0, valB - 2^i >= 0
		
		if(a0 or b0) then c = c + 2^i end
		if(a0) then valA = valA - 2^i end
		if(b0) then valB = valB - 2^i end
		if(valA == 0) then break end
		if(valB == 0) then break end
	
	end
	
	bit.writeBytexy(bit.registerX(regA), bit.registerY(regA), c)

end

function bit.notRegisters(reg) -- bitwise not the values of regA and regB

	local regA, regB = math.floor(reg / 16), reg % 16
	
	local valA, valB = bit.readBytexy(bit.registerX(regA), bit.registerY(regA)), bit.readBytexy(bit.registerX(regB), bit.registerY(regB))
	
	bit.writeBytexy(bit.registerX(regA), bit.registerY(regA), (-1 - valA))
	bit.writeBytexy(bit.registerX(regB), bit.registerY(regB), (-1 - valB))

end

function bit.xorRegisters(reg) -- bitwise xor the value of regB to regA

	local regA, regB = math.floor(reg / 16), reg % 16
	
	local valA, valB = bit.readBytexy(bit.registerX(regA), bit.registerY(regA)), bit.readBytexy(bit.registerX(regB), bit.registerY(regB))
	
	local c = 0
	
	for i=7,0,-1 do
	
		local a0, b0 = valA - 2^i >= 0, valB - 2^i >= 0
		
		if((a0 or b0) and not(a0 and b0)) then c = c + 2^i end
		if(a0) then valA = valA - 2^i end
		if(b0) then valB = valB - 2^i end
		if(valA == 0) then break end
		if(valB == 0) then break end
	
	end
	
	bit.writeBytexy(bit.registerX(regA), bit.registerY(regA), c)

end


-- system functions

function bit.registerX(reg)

	reg = reg % 8

	if(reg < 4) then return 52 
	else return 75
	end

end

function bit.registerY(reg)
	
	return 67 + ((reg % 4) * 2)

end


function bit.readByte(addr)

	return bit.readBytexy(math.floor(addr / bit.MEMORY_HEIGHT) * 2, addr % bit.MEMORY_HEIGHT)

end

function bit.readBytexy(x, y)

	local pos1, pos2 = objectat(x, y), objectat((x + 1), y)

	return bit.decode(pos1) * 16 + bit.decode(pos2)

end

function bit.writeByte(addr, val)

	bit.writeBytexy(math.floor(addr / bit.MEMORY_HEIGHT) * 2, addr % bit.MEMORY_HEIGHT, val)

end

function bit.writeBytexy(x, y, val)

	local val1,val2 = math.floor(val / 16), val % 16
	
	local team1, team2 = 1, 1 -- T as default
	
	if(val1 > 8) then
		val1 = val1 % 8
		team1 = 2
	elseif(val1 == 8) then
		team1 = 2
	end
	
	if(val2 > 8) then 
		val2 = val2 % 8
		team2 = 2
	elseif(val2 == 8) then
		team2 = 2
	end
	
	parse("killobject "..objectat(x, y))
	if(val1 > 0) then
		parse("spawnobject "..bit.encode(val1).." "..x.." "..y.." 0 0 "..team1)
	end
	
	parse("killobject "..objectat((x + 1), y))
	if(val2 > 0) then
		parse("spawnobject "..bit.encode(val2).." "..(x + 1).." "..y.." 0 0 "..team2)
	end

end

function bit.decode(id)

	local objTeam, objType = 0, 0;

	if(object(id, "team")) then objTeam = object(id, "team") end
	if(object(id, "type")) then objType = object(id, "type") end

	objTeam = math.min(2, math.max(0, objTeam - 1)) -- 0 for Ts/neural, 1 for CTs/VIPs
	
	local val = 0
	
	if(objType == 2) then val = 1		-- Barbed wire
	elseif(objType == 1) then val = 2	-- Barricade
	elseif(objType == 3) then val = 3	-- Wall I
	elseif(objType == 4) then val = 4	-- Wall II
	elseif(objType == 5) then val = 5	-- Wall III
	elseif(objType == 8) then val = 6	-- Turret
	elseif(objType == 9) then val = 7	-- Supply
	end
	
	return val + objTeam * 8 -- highest bit determined by team

end

function bit.encode(val)

	if(val == 1) then return 2
	elseif(val == 2) then return 1
	elseif(val == 3) then return 3
	elseif(val == 4) then return 4
	elseif(val == 5) then return 5
	elseif(val == 6) then return 8
	elseif(val == 7) then return 9
	end
	
	return 6 -- gate field as building for 16

end


--hooks & timers


addhook("trigger","bit.handleTrigger")
addhook("serveraction","bit.handleFKeys")

timer(15000,"parse","msg 8bit 4K machine by Hador")
timer(15000,"parse",'msg "[F2] for execute, [F3] for abort, [F4] for value input"')